/*
 * ====================================================================
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2001 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 * if any, must include the following acknowledgment:
 * "This product includes software developed by the
 * Apache Software Foundation (http://www.apache.org/)."
 * Alternately, this acknowledgment may appear in the software itself,
 * if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Apache" and "Apache Software Foundation" and
 * "Apache JMeter" must not be used to endorse or promote products
 * derived from this software without prior written permission. For
 * written permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 * "Apache JMeter", nor may "Apache" appear in their name, without
 * prior written permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */
package org.apache.jmeter.protocol.http.config;

import java.util.*;
import java.net.*;

import org.apache.jmeter.config.*;
import org.apache.jmeter.util.JMeterUtils;
import java.io.*;

/**
 *  Title: Apache JMeter Description: Copyright: Copyright (c) 2000 Company:
 *  Apache Foundation
 *
 *  This class defines the default properties for an HTTP Request element. So,
 *  if you do not specify a value for an HTTP Request field, JMeter uses
 *  the value you specify here.
 *
 *@author     Michael Stover
 *@created    $Date: 2002/02/23 01:21:06 $
 *@version    $Revision: 1.27 $
 */

public class UrlConfig extends AbstractConfigElement implements Serializable
{
	/**
	 *  Description of the Field
	 */
	public final static String DOMAIN = "domain";
	/**
	 *  Description of the Field
	 */
	public final static String PORT = "port";
	/**
	 *  Description of the Field
	 */
	public final static String PATH = "path";
	/**
	 *  Description of the Field
	 */
	public final static String METHOD = "method";
	/**
	 *  Description of the Field
	 */
	public final static String ARGUMENTS = "arguments";
	/**
	 *  Description of the Field
	 */
	public final static String POST = "POST";
	/**
	 *  Description of the Field
	 */
	public final static String PROTOCOL = "PROTOCOL";
	/**
	 *  Description of the Field
	 */
	public final static String GET = "GET";

	/**
	 *  Constructor for the UrlConfig object
	 */
	public UrlConfig()
	{
		this.putProperty(ARGUMENTS, new Arguments());
	}

	/**
	 *  Sets the Method attribute of the UrlConfig object
	 *
	 *@param  meth  The new Method value
	 */
	public void setMethod(String meth)
	{
		properties.put(METHOD, meth);
	}

	/**
	 *  Sets the Protocol attribute of the UrlConfig object
	 *
	 *@param  protocol  The new Protocol value
	 */
	public void setProtocol(String protocol)
	{
		properties.put(PROTOCOL, protocol);
	}

	public void removeArguments()
	{
		properties.put(ARGUMENTS,new Arguments());
	}

	/**
	 *  Sets the Port attribute of the UrlConfig object
	 *
	 *@param  port  The new Port value
	 */
	public void setPort(int port)
	{
		properties.put(PORT, new Integer(port));
		if (port == 0)
		{
			properties.remove(PORT);
		}
	}

	/**
	 *  Sets the Path attribute of the UrlConfig object
	 *
	 *@param  path  The new Path value
	 */
	public void setPath(String path)
	{
		int index = path.indexOf("?");
		if (index >= 0)
		{
			String queryString = path.substring(index + 1);
			properties.put(PATH, path.substring(0, index));
			parseArguments(queryString);
		}
		else
		{
			properties.put(PATH, path);
		}
	}

	public void addArgument(String name,String value)
	{
		Arguments args = this.getArguments();
		args.addArgument(name,value);
	}

	/**
	 *  Sets the Domain attribute of the UrlConfig object
	 *
	 *@param  domain  The new Domain value
	 */
	public void setDomain(String domain)
	{
		properties.put(DOMAIN, domain);
	}

	/**
	 *  !ToDoo (Method description)
	 *
	 *@return                            !ToDo (Return description)
	 *@exception  MalformedURLException  !ToDo (Exception description)
	 */
	public URL getUrl() throws MalformedURLException
	{
		String pathAndQuery = null;

		if (this.getMethod().equals(this.GET) && getQueryString().length() > 0)
		{
			pathAndQuery = this.getPath() + "?" + getQueryString();
		}
		else
		{
			pathAndQuery = this.getPath();
		}

		if (!pathAndQuery.startsWith("/"))
		{
			pathAndQuery = "/" + pathAndQuery;
		}

		if (getPort() == 0)
		{
			return new URL(getProtocol(), getDomain(), pathAndQuery);
		}
		else
		{
			return new URL(getProtocol(), (String) properties.get(DOMAIN), getPort(),
					pathAndQuery);
		}
	}

	/**
	 *  Gets the Port attribute of the UrlConfig object
	 *
	 *@return    The Port value
	 */
	public int getPort()
	{
		Object port = properties.get(PORT);
		if (port == null)
		{
			properties.remove(PORT);
			return 0;
		}
		else if (port instanceof Integer)
		{
			return ((Integer) port).intValue();
		}
		else if (port instanceof String)
		{
			Integer intPort = new Integer((String) port);
			properties.put(PORT, intPort);
			return intPort.intValue();
		}
		return 0;
	}

	/**
	 *  Gets the Protocol attribute of the UrlConfig object
	 *
	 *@return    The Protocol value
	 */
	public String getProtocol()
	{
		String protocol = (String) getProperty(PROTOCOL);
		return protocol;
	}

	/**
	 *  Gets the Path attribute of the UrlConfig object
	 *
	 *@return the path if it exists; otherwise, return a single forward slash
	 */
	public String getPath()
	{
		String path = (String)properties.get(PATH);
		if (path == null)
		{
			path = "/";
		}

		return path;
	}

	/**
	 *  Gets the Domain attribute of the UrlConfig object
	 *
	 *@return    The Domain value
	 */
	public String getDomain()
	{
		return (String) properties.get(DOMAIN);
	}

	/**
	 *  Gets the Method attribute of the UrlConfig object
	 *
	 *@return    The Method value
	 */
	public String getMethod()
	{
		return (String) properties.get(METHOD);
	}

	/**
	 *  Gets the Arguments attribute of the UrlConfig object
	 *
	 *@return    The Arguments value
	 */
	public Arguments getArguments()
	{
		return (Arguments) getProperty(this.ARGUMENTS);
	}

	/**
	 *  Gets the QueryString attribute of the UrlConfig object
	 *
	 *@return    The QueryString value
	 */
	public String getQueryString()
	{
		StringBuffer buf = new StringBuffer();
		Iterator iter = getArguments().iterator();
		boolean first = true;
		while (iter.hasNext())
		{
			Argument item = (Argument) iter.next();
			if (!first)
			{
				buf.append("&");
			}
			else
			{
				first = false;
			}
			buf.append(URLEncoder.encode(item.getName()) + "=" + URLEncoder.encode((String)item.getValue()));
		}
		return buf.toString();
	}

	/**
	 *  Gets the GuiClass attribute of the UrlConfig object
	 *
	 *@return    The GuiClass value
	 */
	public Class getGuiClass()
	{
		return org.apache.jmeter.protocol.http.config.gui.UrlConfigGui.class;
	}

	/**
	 *  Gets the ClassLabel attribute of the UrlConfig object
	 *
	 *@return    The ClassLabel value
	 */
	public String getClassLabel()
	{
		return JMeterUtils.getResString("url_config_title");
	}

	/**
	 *  Gets the Complete attribute of the UrlConfig object
	 *
	 *@return    The Complete value
	 */
	public boolean isComplete()
	{
		boolean isGood = false;
		try
		{
			this.getUrl();
			isGood = true;
		}
		catch (Exception ex)
		{
		}
		return isGood;
	}

	/**
	 *  Description of the Method
	 *
	 *@return    Description of the Returned Value
	 */
	public boolean expectsModification()
	{
		return false;
	}

	/**
	 *  !ToDo (Method description)
	 *
	 *@return    !ToDo (Return description)
	 */
	public Object clone()
	{
		UrlConfig newConfig = new UrlConfig();
		configureClone(newConfig);
		return newConfig;
	}

	/************************************************************
	 *  !ToDo (Method description)
	 *
	 *@param  newConfig  !ToDo (Parameter description)
	 ***********************************************************/
	protected void configureClone(AbstractConfigElement newConfig)
	{
		super.configureClone(newConfig);
		newConfig.putProperty(this.ARGUMENTS,this.getArguments().clone());
	}

	/**
	 *  This allows config elements to combine and give a "layered" effect. for
	 *  example, say there are two HTTPConfigElements, which have properties for
	 *  domain, path, method, and parameters. If element A has everything filled
	 *  in, but null for domain, and element B is added, which has only domain
	 *  filled in, then after adding B to A, A will have the domain from B. If A
	 *  already had a domain, then the correct behavior is for A to ignore the
	 *  addition of element B.
	 *
	 *@param  config  !ToDo
	 */
	public void addConfigElement(ConfigElement config)
	{
		if (config instanceof UrlConfig)
		{
			updatePropertyIfAbsent((UrlConfig) config);
		}
	}

	/**
	 * This method allows a proxy server to send over the raw text from a browser's
	 * output stream to be parsed and stored correctly into the UrlConfig object.
	 */
	public void parseArguments(String queryString)
	{
		String[] args = JMeterUtils.split(queryString, "&");
		for (int i = 0; i < args.length; i++)
		{
			int index = args[i].indexOf("=");
			if(index > -1)
			{
				String name = args[i].substring(0,index);
				String value = args[i].substring(index + 1);

				// In JDK 1.2, the decode() method has a throws clause:
				// "throws Exception". In JDK 1.3, the method does not have
				// a throws clause. So, in order to be JDK 1.2 compliant,
				// we need to add a try/catch around the method call.
				try
				{
					addArgument(name, URLDecoder.decode(value));
				}
				catch (Exception e)
				{
					System.out.println("UrlConfig:parseArguments(): Unable to parse argument=[" + value + "]");
					System.out.println("UrlConfig:parseArguments(): queryString=[" + queryString + "]");
				}
			}
		}
	}

	public static UrlConfig createConfig(String contentType)
	{
		if(contentType != null && contentType.startsWith(MultipartUrlConfig.MULTIPART_FORM))
		{
			return new MultipartUrlConfig(contentType.substring(contentType.indexOf("oundary=")+8));
		}
		else
		{
			return new UrlConfig();
		}
	}

}


