/*
 * ====================================================================
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2001 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 * if any, must include the following acknowledgment:
 * "This product includes software developed by the
 * Apache Software Foundation (http://www.apache.org/)."
 * Alternately, this acknowledgment may appear in the software itself,
 * if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Apache" and "Apache Software Foundation" and
 * "Apache JMeter" must not be used to endorse or promote products
 * derived from this software without prior written permission. For
 * written permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 * "Apache JMeter", nor may "Apache" appear in their name, without
 * prior written permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */
 package org.apache.jmeter.visualizers;

import java.awt.*;
import java.util.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.tree.*;
import javax.swing.event.*;
import org.apache.jmeter.reporters.ResultCollectorIndGraph;
import org.apache.jmeter.samplers.SampleResult;
import org.apache.jmeter.samplers.Sampler;
import org.apache.jmeter.gui.*;
import org.apache.jmeter.gui.tree.*;
import org.apache.jmeter.util.ColorHelper;

import org.apache.log4j.*;
import junit.framework.*;

import org.apache.jmeter.protocol.http.sampler.HTTPSamplerFull;
/**
 * Allows the tester to view the response time of the samples in a graph.
 *
 * @author     Khor Soon Hin
 * @created    2001/08/31
 * @version    1.0
 */
public class IndGraphTreeVisualizer extends TreeVisualizer
{
	private static Category catClass =
		Category.getInstance(IndGraphTreeVisualizer.class.getName());

	protected static final Color INITIAL_COLOR = Color.black;
	protected static final String SAMPLES = "IndGraphTreeVisualizer.SAMPLES";
	protected static final String TOTAL_MAX = "IndGraphTreeVisualizer.TOTAL_MAX";
	protected static final String SEPARATOR = "|";
	protected static final int COLOR_INCREMENT = 100;
	protected static final int PLOT_X_WIDTH = 10;

	// stores an entry for each type of page sampled
	// e.g. an entry will exist for :
	// 1. http://www.abc.com/index.html
	// 2. http://www.abc.com/index.html?id=me
	// 3. http://www.abc.com/first.html
	// currently there 1 and 2 are not differentiated but there might be
	// a need in the future
	protected Hashtable ht;
	protected Hashtable colorMapper;
	protected Hashtable maxHash;
	protected ResultCollectorIndGraph model;

	public IndGraphTreeVisualizer()
	{
		catClass.debug("Start : IndGraphTreeVisualizer1");
		ht = new Hashtable();
		colorMapper = new Hashtable();
		maxHash = new Hashtable();
		catClass.debug("End : IndGraphTreeVisualizer1");
	}

	/**
	 * Set the reporter type this visualizer will work with
	 *
	 * @param model	reporter instance
	 */
	public void setModel(Object model)
	{
		catClass.debug("Start : setModel1");
		this.model = (ResultCollectorIndGraph)model;
		this.model.setListener(this);
		init();
		catClass.debug("End : setModel1");
	}

	/**
	 * Initialize this visualizer
	 */
	protected void init()
	{
		catClass.debug("Start : init1");
		SampleResult rootSampleResult = new SampleResult();
		// JGraphicsPanel is a inner class
		resultPanel = new JGraphicsPanel(gridBag);
		resultPanel.setPreferredSize(new Dimension(2000, 800));
		resultPane = new JScrollPane(resultPanel);
		rootSampleResult.putValue(SampleResult.DISPLAY_NAME, "Root");
		root = new DefaultMutableTreeNode(rootSampleResult);
		treeModel = new DefaultTreeModel(root);
		jTree = new JTree(treeModel);
		jTree.getSelectionModel().setSelectionMode(
			TreeSelectionModel.SINGLE_TREE_SELECTION);
		jTree.setCellRenderer(new ColorRenderer());
		jTree.addTreeSelectionListener(this);
		treePane = new JScrollPane(jTree);
		gridBag = new GridBagLayout();
		gbc = new GridBagConstraints();
		treeSplitPane = new JSplitPane(JSplitPane.VERTICAL_SPLIT,
			treePane, resultPanel);
		add(treeSplitPane);
		catClass.debug("End : init1");
	}

	/**
	 * Update the visualizer with new data
	 */
	public void updateGui()
	{
		catClass.debug("Start : updateGui1");
System.out.println("IndGraphTree : updateGui");
		// black is used for the total load time inclusive of all components
		Color currColor = INITIAL_COLOR;
		DefaultMutableTreeNode currNode = null;
		DefaultMutableTreeNode node = null;
		SampleResult nodeSampleResult = null;
		SampleResult res = null;
		String sampleName = null;		// name of the main page
		String sampleNameChild = null;	// name of the component in the main page
		while(model.next() && (res = model.getCurrentSample()) != null)
		{
			// don't use SampleResult.SAMPLE_LABEL 'cos that is only set
			// if the url exist
			sampleName = (String)res.getValue(SampleResult.DISPLAY_NAME);
			// check if there is an entry for this sampled page
			if(catClass.isDebugEnabled())
			{
				catClass.debug("updateGui1 : Looking for sample entry - "
					+ sampleName);
			}
			node = (DefaultMutableTreeNode)ht.get(sampleName);
			if(catClass.isDebugEnabled())
			{
				catClass.debug("updateGui1 : node of sample entry - " + node);
			}
			if(node != null)
			{
				// update the entry
				nodeSampleResult = (SampleResult)node.getUserObject();
				java.util.List list = (java.util.List)nodeSampleResult.getValue(SAMPLES);
				// just keep the time instead of the whole sample object
				// to minimize memory usage
				list.add(res);
				long max = getTotalMax(nodeSampleResult, res);
				nodeSampleResult.putValue(TOTAL_MAX, new Long(max));
			}
			else
			{
				// create a new entry
				currNode = new DefaultMutableTreeNode(res);
				treeModel.insertNodeInto(currNode, root, childIndex++);
				java.util.List list = new java.util.LinkedList();
				// just keep the time instead of the whole sample object
				// to minimize memory usage
				long time = getTotalLoadTime(res);
				list.add(res);
				res.putValue(SAMPLES, list);
				// max must not be zero
				if(time == (long)0)
				{
					time = 1;
				}
				res.putValue(TOTAL_MAX, new Long(time));
				ht.put(sampleName, currNode);
				currColor = ColorHelper.changeColorCyclicIncrement(currColor,
					COLOR_INCREMENT);
				colorMapper.put(sampleName, currColor);
			}
			ArrayList arrayList =
				(ArrayList)res.getValue(SampleResult.RESULT_LIST);
			if(arrayList != null)
			{
				Iterator iter = arrayList.iterator();
				int leafIndex = 0;
				while(iter.hasNext())
				{
					SampleResult child = (SampleResult)iter.next();
					sampleNameChild = (String)
						child.getValue(SampleResult.DISPLAY_NAME);
					String combinedName = sampleName + SEPARATOR +
						sampleNameChild;
					if(catClass.isDebugEnabled())
					{
						catClass.debug(
							"updateGui1 : "
							+ "Looking for child sample entry - "
							+ combinedName);
					}
					node = (DefaultMutableTreeNode)
						ht.get(combinedName);
					if(catClass.isDebugEnabled())
					{
						catClass.debug("updateGui1 : node of sample entry - " 						+ node);
					}
					if(node != null)
					{
						// update entry
						nodeSampleResult =
							(SampleResult)node.getUserObject();
						java.util.List list =
							(java.util.List)nodeSampleResult.getValue(SAMPLES);
						// just keep the time instead of the whole sample
						// object to minimize memory usage
						list.add(child);
					}
					else
					{
						// create new entry
						DefaultMutableTreeNode leafNode =
							new DefaultMutableTreeNode(child);
						treeModel.insertNodeInto(leafNode, currNode,
							leafIndex++);
						java.util.List list = new java.util.LinkedList();
						// just keep the time instead of the whole sample
						// object to minimize memory usage
						list.add(child);
						child.putValue(SAMPLES, list);
						ht.put(combinedName, leafNode);
						currColor =
							ColorHelper.changeColorCyclicIncrement(
								currColor, COLOR_INCREMENT);
						colorMapper.put(combinedName, currColor);
					}
				}
			}
		}
		catClass.debug("End : updateGui1");
	}

	/**
	 * Given a <code>TreePath</code> returns a string which contains all
	 * the path concatenated but separated with a specified <code>String</code>.
	 *
	 * @param treePath	the <code>TreePath</code> to be converted to
	 * 				<code>String</code>
	 * @param separator	separate each path with this
	 */
	protected String getTreePath(TreePath treePath, String separator)
	{
		catClass.debug("Start : getTreePath1");
		if(catClass.isDebugEnabled())
		{
			catClass.debug("getTreePath1 : treePath - " + treePath);
		}
		StringBuffer treePathStrBuff = new StringBuffer();
		String treePathStr = null;
		Object[] objs = treePath.getPath();
		if(objs != null)
		{
			if(catClass.isDebugEnabled())
			{
				catClass.debug("total path - " + objs.length);
			}
			// don't include the first path 'cos its root
			for(int i = 1; i < objs.length-1; i++)
			{
				treePathStrBuff.append(objs[i].toString());
				treePathStrBuff.append(separator);
			}
			treePathStrBuff.append(objs[objs.length-1].toString());
			treePathStr = treePathStrBuff.toString();
		}
		if(catClass.isDebugEnabled())
		{
			catClass.debug("getTreePath1 : treePath - " + treePathStr);
		}
		catClass.debug("End : getTreePath1");
		return treePathStr;
	}

	public void valueChanged(TreeSelectionEvent e)
	{
		catClass.debug("Start : valueChanged1");
		int row = jTree.getMinSelectionRow();
		TreePath treePath = jTree.getPathForRow(row);
		if(treePath != null)
		{
			String sampleName = getTreePath(treePath, SEPARATOR);
			DefaultMutableTreeNode node =
				(DefaultMutableTreeNode)ht.get(sampleName);
			if(catClass.isDebugEnabled())
			{
				catClass.debug("valueChanged1 : Sample name - " + sampleName);
				catClass.debug("valueChanged1 : Selected node - " + node);
			}
			if(node != null)
			{
				SampleResult res = (SampleResult)node.getUserObject();
				if(catClass.isDebugEnabled())
				{
					catClass.debug("valueChanged1 : Sample result - " + res);
				}
				if(res != null)
				{
					java.util.List list =
						(java.util.List)res.getValue(SAMPLES);
					Graphics g = resultPanel.getGraphics();
					Dimension d = resultPanel.getSize();
					g.clearRect(0, 0, d.width, d.height);
					synchronized(list)
					{
						Iterator iter = list.iterator();
						for (int i = 0; iter.hasNext(); i++)
						{
							SampleResult s = (SampleResult)iter.next();
							drawSample(i, s, treePath ,g);
						}
					}
				}
			}
		}
		catClass.debug("End : valueChanged1");
	}

	/**
	 * Plots a sample.  Each sample can contain other samples.  E.g. a.html may
	 * contain b.jpg.  In this case the sample for a.html will contain sample
	 * for b.jpg.  A <code>JTree</code> will be used to reflect this relationship to
	 * the user.  a.html sample node will contain b.jpg sample as its child.
	 * If a.html is sampled x times, then all the samples
	 * of a.html will be stored in the sample of a.html under SAMPLES key.
	 * Similarly, b.jpg will be sampled x times since it's in a.html.  It's samples
	 * will be store in the samples of b.jpg under SAMPLES.key.
	 * In order for the program to group all samples of a.html under the same node
	 * a hashtable is used to store the sample name with it related node.  Since
	 * b.jpg might occur in other htmls, the b.jpg samples under a.html samples will
	 * be store as a.html:b.jpg.  This way subsequent a.html samples can be stored
	 * in the same node by look for the node corresponding to the name from the
	 * hashtable.  The <code>treePath</code> passed in is required because sometimes
	 * the user selects a node which is component of another e.g. b.jpg (of a.html)
	 * In this case,
	 *
	 * @param i			the number of samples so far (used as x-axis)
	 * @param oneSample	the sample to be plot
	 * @param treePath	treePath of this sample
	 * @param g			where the plotting should be done
	 */
	protected void drawSample(int i, SampleResult oneSample, TreePath treePath,
		Graphics g)
	{
		catClass.debug("Start : drawSample1");
		int x = i*PLOT_X_WIDTH;
		int prevX = x-PLOT_X_WIDTH;
		if(prevX < 0)
		{
			prevX = 0;
		}
		Color currColor = null;
		DefaultMutableTreeNode node = null;
		Dimension d = resultPanel.getSize();
		int lastLevel = 0;
		int lastLevely = 0;
		long last = (long)0;
		java.util.List list = null;
		Long maxLong = null;
		long max = (long)0;
		SampleResult lastSample = null;
		SampleResult res = null;
		// if the sample has other components then resultList won't be null
		ArrayList resultList =
			(ArrayList)oneSample.getValue(SampleResult.RESULT_LIST);
		StringBuffer sampleNameBuff = new StringBuffer();
		sampleNameBuff.append(getTreePath(treePath, SEPARATOR));
		String sampleName = sampleNameBuff.toString();
		long totalTime = getTotalLoadTime(oneSample);
		if(catClass.isDebugEnabled())
		{
			catClass.debug("drawSample1 : Total time - " + totalTime);
			catClass.debug("drawSample1 : Looking up samples for - " +
				sampleName);
		}
		// get the max value of the graph so far
		// get max value from main node
		// component at 0 is root while component at 1 is the main node
		String mainComponent = treePath.getPathComponent(1).toString();
		DefaultMutableTreeNode mainNode =
			(DefaultMutableTreeNode)ht.get(mainComponent);
		if(catClass.isDebugEnabled())
		{
			catClass.debug("drawSample1 : mainComponent - " + mainComponent);
			catClass.debug("drawSample1 : mainNode for mainComponent - " +
				mainNode);
		}
//		SampleResult mainSampleResult = (SampleResult)mainNode.getUserObject();
//		max = getTotalLoadTime(mainSampleResult);
		maxLong = (Long)maxHash.get(mainComponent);
		// if null is obtained from maxHash, then this is the first sample for
		// this component, so store is as max
		if(maxLong == null)
		{
			maxLong = new Long(totalTime);
			maxHash.put(mainComponent, maxLong);
		}
		max = maxLong.longValue();
		if(catClass.isDebugEnabled())
		{
//			catClass.debug("drawSample1 : Component storing max - " +
//				mainNode);
			catClass.debug("drawSample1 : Max - " + max);
		}
		if(totalTime > max)
		{
//			mainSampleResult.putValue(SampleResult.TOTAL_TIME,
//				new Long(totalTime));
			maxHash.put(mainComponent, new Long(totalTime));
		}
		// get the last point drawn for this graph
		node = (DefaultMutableTreeNode)ht.get(sampleName);
		res = (SampleResult)node.getUserObject();
		list = (java.util.List)res.getValue(SAMPLES);
		if(i != 0)
		{
			lastSample = (SampleResult)list.get(i-1);
		}
		else
		{
			lastSample = (SampleResult)list.get(i);
		}
		last = getTotalLoadTime(lastSample);
		int data = (int)(totalTime * d.height / max);
		int datay = (int)(last * d.height / max);
		g.setColor(INITIAL_COLOR);
		g.drawLine((prevX)%d.width, d.height - datay,
				x%d.width, d.height - data);

		// plot the time of the page itself without all its components
		if(catClass.isDebugEnabled())
		{
			catClass.debug("drawSample1 : main page load time - "
				+ oneSample.getTime());
		}
		// get the last point drawn for this graph
		last = lastSample.getTime();
		data = (int)(oneSample.getTime()
			* d.height / max);
		datay = (int)(last * d.height / max);
		currColor = (Color)colorMapper.get(sampleName);
		g.setColor(currColor);
		g.drawLine((prevX)%d.width, d.height - datay, x%d.width, d.height - data);

		lastLevel += data;
		lastLevely += datay;
		// plot the times of the total times components
		if(resultList != null)
		{
			Iterator iter = resultList.iterator();
			while(iter.hasNext())
			{
				SampleResult componentRes = (SampleResult)iter.next();
				String sampleNameChild = (String)
					componentRes.getValue(SampleResult.DISPLAY_NAME);
				String combinedName = sampleName + SEPARATOR + sampleNameChild;
				if(catClass.isDebugEnabled())
				{
					catClass.debug("drawSample1 : componentRes - " +
						combinedName + " loading time - " +
							componentRes.getTime());
				}
				node = (DefaultMutableTreeNode)ht.get(combinedName);
				res = (SampleResult)node.getUserObject();
				list = (java.util.List)res.getValue(SAMPLES);
				if(i != 0)
				{
					lastSample = (SampleResult)list.get(i-1);
				}
				else
				{
					lastSample = (SampleResult)list.get(i);
				}
				last = lastSample.getTime();
				datay = (int)(last * d.height / max);
				datay += lastLevely;
				data = (int)(componentRes.getTime()
					* d.height / max);
				data += lastLevel;
				if(catClass.isDebugEnabled())
				{
					catClass.debug("drawSample1 : componentRes data - " +
						data);
					catClass.debug("drawSample1 : componentRes datay - " +
						datay);
				}
				currColor = (Color)colorMapper.get(combinedName);
				g.setColor(currColor);
				g.drawLine((prevX)%d.width,
						d.height - datay, x%d.width,
						d.height - data);
				lastLevel = data;
				lastLevely = datay;
			}
		}
		resultPanel.repaint();
		catClass.debug("End : drawSample1");
	}

	/**
	 * Returns the description of this visualizer
	 *
	 * @return    description of this visualizer
	 */
	public String toString()
	{
		String desc = "Shows the results of sampling in a graph";
		if(catClass.isDebugEnabled())
		{
			catClass.debug("toString1 : Returning description - " + desc);
		}
		return desc;
	}

	/**
	 * Compares two <code>SampleResult</code> total load time (inclusive of all
	 * components) and return the greater
	 * of the two
	 *
	 * @param sample1		the first sample
	 * @param sample2		the second sample
	 */
	protected long getTotalMax(SampleResult sample1, SampleResult sample2)
	{
		long max = (long)0;
		long sample1Time = ((Long)sample1.getValue(TOTAL_MAX)).longValue();
		long sample2Time = getTotalLoadTime(sample2);
		if(sample1Time < sample2Time)
		{
			max = sample2Time;
		}
		else
		{
			max = sample1Time;
		}
		if(catClass.isDebugEnabled())
		{
			catClass.debug("getTotalMax1 : Returning max - " + max);
		}
		return max;
	}

	/**
	 * !!!Refactor this into HttpSampleResult
	 * Return the total load time inclusive of component load times if available
	 *
	 * @param res	the <code>SampleResult</code> whose load time is required
	 * @return 		the total load time of the sample and all its componenets
	 */
	protected long getTotalLoadTime(SampleResult res)
	{
		long time = res.getTime();
		Long timeLong = (Long)res.getValue(SampleResult.TOTAL_TIME);
		if(timeLong != null)
		{
			time = timeLong.longValue();
		}
		return time;
	}

	public void clear()
	{
		catClass.debug("Start : clear1");
		super.clear();
//		Graphics g = resultPanel.getGraphics();
//		Dimension d = resultPanel.getSize();
//		g.clearRect(0, 0, d.width, d.height);
//		resultPanel.revalidate();
		ht.clear();
		colorMapper.clear();
		maxHash.clear();
		catClass.debug("End : clear1");
	}

	private class ColorRenderer extends DefaultTreeCellRenderer
	{
		public Component getTreeCellRendererComponent(
			JTree tree, Object value, boolean sel, boolean expanded,
			boolean leaf, int row, boolean hasFocus)
		{
			super.getTreeCellRendererComponent(
				tree, value, sel,
				expanded, leaf, row,
				hasFocus);
			catClass.debug("Start : getTreeCellRendererComponent1");
			TreePath treePath = jTree.getPathForRow(row);
			if(catClass.isDebugEnabled())
			{
				catClass.debug("getTreeCellRendererComponent1 : row - " + row);
				catClass.debug("getTreeCellRendererComponent1 : treePath - " +
					treePath);
			}
			String sampleName = null;
			Color currColor = null;
			// jTree.getPathForRow(row) can be null at certain time
			if(treePath != null)
			{
				sampleName = getTreePath(treePath, SEPARATOR);
				currColor = (Color)colorMapper.get(sampleName);
				setForeground(currColor);
				if(catClass.isDebugEnabled())
				{
					catClass.debug("getTreeCellRendererComponent1 : sample - " 						+ sampleName);
					catClass.debug("getTreeCellRendererComponent1 : color - " +
						currColor);
				}
			}
			catClass.debug("End : getTreeCellRendererComponent1");
			return this;
		}
	}

	/**
	 * Class which knows how to redraw the graphs when repainted.
	 */
	private class JGraphicsPanel extends JPanel
	{
		/**
		 * Constructor that accepts a layout manager
		 */
		public JGraphicsPanel(LayoutManager lm)
		{
			super(lm);
		}

		/**
		 * This method is called when the component needs to be redrawn.
		 *
		 * @param  g  Description of Parameter
		 */
		public void paintComponent(Graphics g)
		{
			super.paintComponent(g);
			catClass.debug("Start : paintComponent1");
//			ht.clear();
			Dimension d = this.getSize();
			DefaultMutableTreeNode node = (DefaultMutableTreeNode)jTree.getLastSelectedPathComponent();
			if(catClass.isDebugEnabled())
			{
				catClass.debug("paintComponent1 : selected node - " + node);
			}
			if(node != null)
			{
				Object userObject = node.getUserObject();
				if(catClass.isDebugEnabled())
				{
					catClass.debug("paintComponent1 : node user object - " + node);
				}
				if(userObject != null)
				{
					SampleResult res = (SampleResult)userObject;
					java.util.List list = (java.util.List)res.getValue(SAMPLES);
					// root node although is an instance of SampleResult
					// contains nothing
					if(list != null)
					{
						Graphics resultPanelGraphics = resultPanel.getGraphics();
						int row = jTree.getMinSelectionRow();
						TreePath treePath = jTree.getPathForRow(row);
						synchronized(list)
						{
							Iterator e = list.iterator();
							for (int i = 0; e.hasNext(); i++)
							{
								SampleResult s = (SampleResult)e.next();
								drawSample(i, s, treePath, //resultPanelGraphics);
g);
							}
						}
					}
				}
			}
			catClass.debug("End : paintComponent1");
		}
	}
}
